#pragma GCC diagnostic ignored "-Wsign-compare"
// ROOT
#include "TFile.h"
#include "TTree.h"
#include "TChain.h"
#include "TTreeCache.h"
#include "TTreeCacheUnzip.h"
#include "TTreePerfStats.h"
#include "Math/LorentzVector.h"
#include "Math/GenVector/PtEtaPhiM4D.h"
#include "Math/VectorUtil.h"
// NanoCORE
#include "Nano.h"
#include "tqdm.h"
// STL
#include <iostream>
#include <iomanip>

int main() 
{
    TChain* chain = new TChain("Events");
    chain->Add("example_RunIISummer20UL18.root");

    TFile* output_file = new TFile("output.root", "RECREATE");

    // Disable all branches
    chain->SetBranchStatus("*", 0);
    // Enable selected branches
    std::vector<TString> keep_branch_names = {
        "Electron*", "nElectron",
        "Muon*", "nMuon",
        "Tau*", "nTau",
        "Jet*", "nJet",
        "FatJet*", "nFatJet",
        "GenPart*", "nGenPart",
        "GenJet*", "nGenJet",
        "Generator*",
        "MET*",
        "event*",
        "run*",
        "luminosityBlock*",
        "genWeight*",
        "btagWeight*",
        "LHE*",
        "*Weight*",
        "Flag*",
        "SubJet*",
        "HLT_*",
        "Pileup*",
        "fixedGridRhoFastjetAll"
    };
    for (auto branch_name : keep_branch_names)
    {
        chain->SetBranchStatus(branch_name, 1);
    }

    TTree* output_tree = (TTree*)chain->CloneTree(0);

    // TList* runs_trees = new TList();
    // TList* lumi_trees = new TList();

    std::vector<LorentzVector> veto_lep_p4s;
    std::vector<LorentzVector> tight_lep_p4s;

    // File loop
    int n_events_parsed = 0;
    int n_events_tchain = chain->GetEntries();
    TFile* current_file = 0;
    tqdm bar;
    TObjArray* file_list = chain->GetListOfFiles();
    TIter file_iter(file_list);
    while ((current_file = (TFile*)file_iter.Next())) 
    {
        TFile* file = TFile::Open(current_file->GetTitle());
        TTree* tree = (TTree*)file->Get("Events");
        tree->SetCacheSize(128*1024*1024);
        tree->SetCacheLearnEntries(100);

        // Disable all branches
        tree->SetBranchStatus("*", 0);
        // Enable selected branches
        for (auto branch_name : keep_branch_names)
        {
            tree->SetBranchStatus(branch_name, 1);
        }
        tree->CopyAddresses(output_tree);

        nt.Init(tree);

        // Store metadata ttrees
        // TTree* runs_tree = ((TTree*)tree->GetCurrentFile()->Get("Runs"))->CloneTree();
        // runs_tree->SetDirectory(0);
        // runs_trees->Add(runs_tree);
        // TTree* lumi_tree = ((TTree*)tree->GetCurrentFile()->Get("LuminosityBlocks"))->CloneTree();
        // lumi_tree->SetDirectory(0);
        // lumi_trees->Add(lumi_tree);

        // Event loop
        for (unsigned int event = 0; event < tree->GetEntriesFast(); ++event) 
        {
            // Load event
            nt.GetEntry(event);

            // Update progress bar
            bar.progress(n_events_parsed, n_events_tchain);
            n_events_parsed++;

            veto_lep_p4s.clear();
            tight_lep_p4s.clear();

            // Select electrons that pass veto or tight lepton ID
            for (unsigned int elec_i = 0; elec_i < nt.nElectron(); ++elec_i)
            {
                LorentzVector elec_p4 = nt.Electron_p4().at(elec_i);
                if (nt.Electron_pt().at(elec_i) <= 10) { continue; }
                if (nt.Electron_cutBased().at(elec_i) < 1) { continue; }
                // Passed veto lepton ID
                veto_lep_p4s.push_back(elec_p4);

                if (nt.Electron_pt().at(elec_i) <= 35) { continue; }
                if (nt.Electron_cutBased().at(elec_i) < 3) { continue; }
                if (fabs(nt.Electron_eta().at(elec_i) + nt.Electron_deltaEtaSC().at(elec_i)) >= 2.5) { continue; }
                if (fabs(nt.Electron_eta().at(elec_i) + nt.Electron_deltaEtaSC().at(elec_i)) >= 1.479)
                {
                    if (fabs(nt.Electron_dz().at(elec_i)) >= 0.2) { continue; }
                    if (fabs(nt.Electron_dxy().at(elec_i)) >= 0.1) { continue; }
                }
                else
                {
                    if (fabs(nt.Electron_dz().at(elec_i)) >= 0.1) { continue; }
                    if (fabs(nt.Electron_dxy().at(elec_i)) >= 0.05) { continue; }
                }
                // Passed tight lepton ID
                tight_lep_p4s.push_back(elec_p4);
            }

            // Select muons that pass veto or tight lepton ID
            for (unsigned int muon_i = 0; muon_i < nt.nMuon(); ++muon_i)
            {
                LorentzVector muon_p4 = nt.Muon_p4().at(muon_i);
                if (!nt.Muon_tightId().at(muon_i)) { continue; }
                if (nt.Muon_pfRelIso04_all().at(muon_i) >= 0.4) { continue; }
                if (nt.Muon_pt().at(muon_i) <= 10) { continue; }
                // Passed veto lepton ID
                veto_lep_p4s.push_back(muon_p4);

                if (nt.Muon_pfRelIso04_all().at(muon_i) >= 0.15) { continue; }
                if (nt.Muon_pt().at(muon_i) <= 26) { continue; }
                if (fabs(nt.Muon_eta().at(muon_i)) >= 2.4) { continue; }
                // Passed tight lepton ID
                tight_lep_p4s.push_back(muon_p4);
            }

            // Require one and only one tight lepton (ignoring taus)
            if (!(veto_lep_p4s.size() == 1 && tight_lep_p4s.size() == 1)) { continue; }

            LorentzVector lep_p4 = tight_lep_p4s.at(0);

            // Require at least 2 AK4 jets (no veto lepton overlap)
            int n_jets = 0;
            for (unsigned int jet_i = 0; jet_i < nt.nJet(); ++jet_i)
            {
                LorentzVector jet_p4 = nt.Jet_p4().at(jet_i);
                bool is_overlap = ROOT::Math::VectorUtil::DeltaR(lep_p4, jet_p4) < 0.4;
                if (!is_overlap && nt.Jet_pt().at(jet_i) > 20)
                {
                    n_jets++;
                }
            }
            if (n_jets < 2) { continue; }

            // Require at least 1 AK8 "fat" jet
            int n_fatjets = 0;
            double hbbjet_score = -999.;
            LorentzVector hbbjet_p4;
            for (unsigned int fatjet_i = 0; fatjet_i < nt.nFatJet(); ++fatjet_i)
            {
                LorentzVector fatjet_p4 = nt.FatJet_p4().at(fatjet_i);
                bool is_overlap = ROOT::Math::VectorUtil::DeltaR(lep_p4, fatjet_p4) < 0.8;
                if (!is_overlap
                    && nt.FatJet_mass().at(fatjet_i) > 50
                    && nt.FatJet_msoftdrop().at(fatjet_i) > 40
                    && nt.FatJet_pt().at(fatjet_i) > 250)
                {
                    // Find candidate with highest Xbb score
                    double pnet_xbb = nt.FatJet_particleNetMD_Xbb().at(fatjet_i);
                    double pnet_qcd = nt.FatJet_particleNetMD_QCD().at(fatjet_i);
                    double xbb_score = pnet_xbb/(pnet_xbb + pnet_qcd);
                    if (xbb_score > hbbjet_score)
                    {
                        hbbjet_score = xbb_score;
                        hbbjet_p4 = fatjet_p4;
                    }
                    n_fatjets++;
                }

            }
            if (n_fatjets < 1) { continue; }

            // Require ST > 800 GeV
            double ST = hbbjet_p4.pt() + lep_p4.pt() + nt.MET_pt();
            if (ST < 800) { continue; }

            tree->GetEntry(event);
            output_tree->Fill();
        }

        // Clean up
        delete file;
    }

    bar.finish();

    // Merge metadata trees
    // TTree* merged_runs_trees = TTree::MergeTrees(runs_trees);
    // merged_runs_trees->SetName("Runs");
    // TTree* merged_lumi_trees = TTree::MergeTrees(lumi_trees);
    // merged_lumi_trees->SetName("LuminosityBlocks");

    // Write output
    output_file->cd();
    // merged_runs_trees->Write();
    // merged_lumi_trees->Write();
    output_file->Write();
    output_file->Close();

    return 0;
}
